from typing import List

from fancy import config as cfg


class MyConfigEmb(cfg.BaseConfig):
    x: int = cfg.Option(type=int)
    y: int = cfg.Option(type=int)
    i: bool = cfg.Option(type=bool)


class MyConfig(cfg.BaseConfig):
    _a: int = cfg.Option(name="a", required=True, description="hi", type=int)
    b: float = cfg.Option(required=False, type=float)
    c: bool = cfg.Option(required=True, type=bool)
    li: List[MyConfigEmb] = cfg.Option(type=cfg.config_list(cfg.config_list(MyConfigEmb)))

    def post_load(self):
        print(self.a)

    @property
    def a(self):
        return self._a + 1


def test_config():
    opt = {
        "a": "3",
        "b": "4.2",
        "c": "true",
        "li": [[{
            "x": 1,
            "y": 2,
            "i": "on",
            "a": "ig"
        },
            {
                "y": 3,
                "i": "off", }
        ]],
        "int_li": ["Y", "N"]
    }

    c = MyConfig(cfg.DictConfigLoader(opt, setter="strict"))
    print(c)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from . import BaseConfigLoader


class ConfigContext(ABC):
    @abstractmethod
    def get_loader(self) -> "BaseConfigLoader": ...



from typing import Union, Type, Callable

from . import boolean, config


def auto_process_typ(typ: Union[Type, Callable]) -> Callable:
    from .. import BaseConfig  # lazy import
    if isinstance(typ, type):
        if issubclass(typ, bool):
            return boolean
        if issubclass(typ, BaseConfig):
            return config(typ)
    if isinstance(typ, Callable):
        return typ
    else:
        raise ValueError("_type must be Callable or subclass of BaseConfig")



from typing import TYPE_CHECKING, Callable, Type, Union


if TYPE_CHECKING:
    from .. import BaseConfig


def config_list(
        _type: Union[Type["BaseConfig"], Callable]) -> Callable:
    from .. import ConfigListStructure

    def _inner(_):
        return ConfigListStructure(_type)
    return _inner



from .option_preprocessing import (
    identical,
    to_float,
    to_integer,
    to_string,
    to_bool,
    to_path,
    to_list
)

from . import attribute_setters

from . import process
from .process import (
    config_list,
    make_boolean
)

from .config_context import ConfigContext
from .config_structure import ConfigStructure
from .option import Option
from .config_list_struct import ConfigListStructure
from .base_config import BaseConfig
from .config_loaders import (
    BaseConfigLoader,
    DictBasedConfigLoader,
    PathBasedConfigLoader,
    YamlConfigLoader,
    DictConfigLoader
)


from .config_loader_factory import ConfigLoaderFactory
from .config_factory import ConfigFactory



import inspect
from abc import ABC
from typing import TYPE_CHECKING, Dict, List, Optional

from . import ConfigStructure, ConfigContext
from . import Option

if TYPE_CHECKING:
    from ..config import BaseConfigLoader


class BaseConfig(ConfigStructure, ConfigContext, ABC):  # TODO more accurate error msg
    _name_mapping: Dict[str, str]
    _all_options: Dict[str, Option]
    _all_required_options: List[Option]
    _loader: Optional['BaseConfigLoader'] = None

    def __init__(self, loader: Optional['BaseConfigLoader'] = None):
        if loader is not None:
            self.load(loader)

    def load(self, loader: 'BaseConfigLoader') -> None:
        self._loader = loader
        loader.load(self)
        for option in self.get_all_required_options():
            if not hasattr(self, option.__name__):
                raise ValueError(f'the missing option {option.name} is required.')
        self.post_load()

    def load_by_context(self, context: ConfigContext, val):
        self.load(context.get_loader().get_sub_loader(val))

    def __getitem__(self, item):
        if isinstance(item, str):
            raise TypeError(f'index must be str, not {type(item)}')
        try:
            return self.__getattribute__(self.get_name_mapping()[item])
        except AttributeError:
            raise KeyError(f'not contains the config named {item}')

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError(f'index must be str, not {type(key)}')
        if key not in self.get_name_mapping().keys():
            raise KeyError(f'not contains the config named {key}, value: {repr(value)}')
        key = self.get_name_mapping()[key]
        self.__setattr__(key, value)

    def get_loader(self) -> 'BaseConfigLoader':
        return self._loader

    def post_load(self):
        pass

    @classmethod
    def get_all_options(cls) -> Dict[str, Option]:
        if "_all_options" not in vars(cls):
            cls._all_options = {name: option for name, option in inspect.getmembers(cls) if isinstance(option, Option)}
        return cls._all_options

    @classmethod
    def get_all_required_options(cls) -> List[Option]:
        if "_all_required_options" not in vars(cls):
            cls._all_required_options = [option for option in cls.get_all_options().values() if option.required]
        return cls._all_required_options

    @classmethod
    def get_name_mapping(cls) -> Dict[str, str]:
        if "_name_mapping" not in vars(cls):
            cls._name_mapping = {option.name: attr_name for attr_name, option in cls.get_all_options().items()}
        return cls._name_mapping

    def __repr__(self):
        return str({
            option.name: getattr(self, option.__name__)
            for option in self.get_all_options().values()
            if option.is_assigned(self)
        })

    def __str__(self):
        return self.__repr__()





import warnings
from typing import Any, Callable, TYPE_CHECKING

from . import ConfigStructure
from .process import auto_process_typ
from ..config import identical

if TYPE_CHECKING:
    from ..config import BaseConfig


class Option:
    _type: Callable[[Any], Any]
    _required: bool
    _nullable: bool
    _default: Any

    _description: str

    _config_name: str = None

    def __init__(
            self,
            required=False,
            nullable=False, default=None,
            type=identical,
            preprocess=identical,
            description="",
            name: str = None
    ):
        self._config_name = name
        self._required = required
        self._nullable = nullable
        self._default = default
        self._description = description

        if preprocess is not identical:
            warnings.warn("preprocess has deprecated. use type to instead.", DeprecationWarning)
            type = preprocess
        self._type = auto_process_typ(type)

    def __get__(self, instance: 'BaseConfig', owner):
        if instance is None:
            return self

        # initialize value
        if self._should_assign_default_value(instance):
            if self._default is None and not self._nullable:
                raise AttributeError("attribute must assign the value before access it.")

            self.__set__(instance, self._default)

        return vars(instance)[self.__name__]

    def __set__(self, instance, raw_value):
        if raw_value is None:
            if not self._nullable:
                raise ValueError('the value should not be none')
            return None

        value = self._type(raw_value)
        if isinstance(value, ConfigStructure):
            value.load_by_context(instance, raw_value)

        vars(instance)[self.__name__] = value

    def __delete__(self, instance):
        del vars(instance)[self.__name__]

    def __set_name__(self, owner, name):
        self.__name__ = name
        if self._config_name is None:
            self._config_name = name

    def is_assigned(self, instance) -> bool:
        return self.__name__ in vars(instance)

    def _should_assign_default_value(self, instance):
        return not self.is_assigned(instance)

    @property
    def name(self) -> str:
        return self._config_name

    @property
    def required(self) -> bool:
        return self._required

    @property
    def description(self) -> str:
        return self._description
